      program tf_pws0MP
c
c Program to perform a zero-lag PWS in the tf-domain
c through the S-Transform.
c 
c Thanks for your interest in using this approach.
c I hope this program is useful to you. Please, note I
c can not take any warranties. I.e., use this program
c on your own risk. 
c Please, report bugs/improvements/... . Also, if you 
c need any help then do not hesitate to contact me.
c (I am preparing a new version which will be faster
c and which can handle larger dimensions.)
c
c Author: Martin Schimmel (schimmel@ictja.csic.es)
c
c Last modification: 
c   13/07/2010 : including kstnm= option
c   09/06/2013 : rm maxtr 
c   12/12/2013 : avoid pg stop in case no evla,evlo,stla,stlo available.
c   30/04/2015 : avoid traces with NaN on input.
c   17/03/2016 : openMP is being used in a simple manner to parallize
c                do loops. tf_pws0.f has been taken from Corr_stack_v03.5
c   13/06/2016 : provide option to write output name and to write
c                sac-hdr variables.
c   02/11/2016 : include option for inverse ST by Stockwell.
c   18/04/2017 : warn if nsamp > nsmax. 
c
c References:
c ===========
c
c This programs performs a PWS (phase weighted stack, Schimmel & Paulssen, 1997)
c in the time-frequency domain employing an S-Transform. This has been
c published in Schimmel & Gallart (2007). There you find also an explanation
c why the S-transform is analytic. To go back into the time domain,
c we use the inverse S-transform from Schimmel & Gallart (2005). This 
c inverse S-transform is also explained in the appendix of 
c Schimmel & Gallart (2007) and further analyzed in Simon et al. (2007)
c and Ventosa et al. (2008). An overview of the ambient noise 
c processing is given in Schimmel et al. (2011). 
c
c  Schimmel M., and H. Paulssen, Noise reduction and
c    detection of weak, coherent signals through phase
c    weighted stacks, GJI, 130, 497-505, 1997.
c  Schimmel M., and J. Gallart, The inverse S Transform in
c    filters with time-frequency localization , IEEE
c    Transactions on Signal Processing, 53 (11), 4417-4422,
c    doi:10.1109/TSP.2005.857065, 2005.
c  Schimmel M., and J. Gallart, Frequency-dependent phase
c    coherence for noise suppression in seismic array data,
c    J. Geophys. Res.,  112, B04303, doi:10.1029/2006JB004680, 2007.
c  Schimmel, M., Stutzmann, E., Gallart, J., Using instantaneous 
c    phase coherence for signal extraction from ambient noise data 
c    at a local to a global scale, Geophys. J. Int., 184, 494-506, 
c    doi: 10.1111/j.1365-246X.2010.04861.x, 2011.
c  Simon, C., Ventosa, S., Schimmel, M., Heldering, A., Dañobeitia JJ, 
c    J. Gallart, A. Manuel, The S-transform and its inverses: side effects 
c    of discretising and filtering, IEEE Transactions on Signal Processing, 
c    55, 4928-4937, doi:10.1109/TSP.2007.897893, 2007.
c  Ventosa, S., Simon, C., Schimmel, M., Dañobeitia JJ, A. Manuel, 
c    The S-transform from a wavelet point of view , IEEE Transactions on 
c    Signal Processing, 56, 2771-2780, doi:10.1109/TSP.2008.917029, 2008.
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c Dimensions & parameters:
c-------------------------
c if changing dimensions remember to change them in all subroutines.
      parameter (nsmax=8192,nfmax=4096)
c     parameter (nsmax=16384,nfmax=8192)
c if changing dimensions remember to change them in all subroutines.
      character*120 filein,tfpws,tls,par
      character*18 par1
      character*8 kstnm
      real sig(nsmax),sig2(nsmax),rsig(nsmax)
      real twpi,pi,stwpi,costapl,f1,f2,t1,t2
      complex ctf(nsmax,nfmax+1)
      complex ctf2(nsmax,nfmax),ctfps(nsmax,nfmax)
      complex carray1(nsmax),carray2(nsmax),cdum,carg
      logical lrm,lkstnm,luser5,luser6,list
      integer nsamp,nsmp,nsmp2,npow

      twpi=8.0*atan(1.0)
      pi=4.0*atan(1.0)
      stwpi=sqrt(twpi)

c INPUTS:
c--------
c  Get number of arguments:
      narg=iargc()
      if (narg.lt.1) then
         write(*,*)' '
         write(*,*)'Dimensions: samples, frequencies:',nsmax,nfmax
         call usage
      endif
 
      call getarg (1,filein)
      if (narg.eq.1.and.filein(1:4).eq.'info') then
       call infooo
       stop
      endif

c DEFAULTS:
C----------
      cycle=2.
      lrm=.false.
      lkstnm=.false.
      wu=2.
      tfpws='tf-pws.sac'
      ntfpws=10
      tls='tl.sac'
      ntls=6
      luser5=.false.
      luser6=.false.
      list=.false.
      
c Get arguments (optional):
c--------------------------
      do iarg=2,narg
        call getarg(iarg,par)
        l=leng(par)
        if (par(1:3).eq.'wu=') then
          par1=par(4:l)
          read(par1,*)wu
        else if (par(1:4).eq.'tls=') then
          tls=par(5:l)
          ntls=l-4
        else if (par(1:6).eq.'tfpws=') then
          tfpws=par(7:l)
          ntfpws=l-6
        else if (par(1:6).eq.'kstnm=') then
          kstnm=par(7:l)
          lkstnm=.true.
        else if (par(1:6).eq.'user5=') then
          par1=par(7:l)
          read(par1,*)user5
          luser5=.true.
        else if (par(1:6).eq.'user6=') then
          par1=par(7:l)
          read(par1,*)user6
          luser6=.true.
        else if (par(1:4).eq.'cyc=') then
          par1=par(5:l)
          read(par1,*)cycle
        else if (par(1:4).eq.'info') then
          call infooo
          stop
        else if (par(1:2).eq.'rm') then
          lrm=.true.
        else if (par(1:3).eq.'ist') then
          list=.true.
        else
          call usage                 
        endif
      enddo

c Zero arrays:
c-------------
c consumes time and is not needed for
c most compilers!
!$OMP PARALLEL DO
!$OMP& PRIVATE (ns,nf)
!$OMP& SHARED (sig,sig2,rsig,ctf,ctfps,ctf2)
      do ns=1,nsmax
        sig(ns)=0.0
        sig2(ns)=0.0
        rsig(ns)=0.0
        do nf=1,nfmax
          ctf(ns,nf)=cmplx(0.0,0.0)
          ctfps(ns,nf)=cmplx(0.0,0.0)
          ctf2(ns,nf)=cmplx(0.0,0.0)
        enddo
      enddo
!$OMP END PARALLEL DO

c read file with data:
c---------------------
      beg=0.0
      nskip=0
      open(1,file=filein,status="old")
      rewind(1)

      itr=0
      mxtr=0
c     do itr=1,maxtr
      do while (mxtr.eq.0)
        itr=itr+1
        read(1,'(A120)',end=80)filein
        call rsac1(filein,sig,nsamp,beg,dt,nsmax,nerr)
        if (itr.eq.1) then
          call getfhv('stla',stla,nerr1)
          call getfhv('stlo',stlo,nerr1)
          call getfhv('evla',evla,nerr1)
          call getfhv('evlo',evlo,nerr1)
        endif

c Some basic checking on the data:
c---------------------------------
        if (nerr.ne.0) then 
          write(*,*)' PROBLEM reading sac data !!!'
          write(*,*)' trace number:',itr
          write(*,*)' file name   :',filein
          if (nsamp-nsmax.le.1) then
             write(*,*)' NOTE number of samples should be smaller'//
     &      ' than',nsmax 
          endif
          stop ' '
        endif
        if (nerr1.ne.0) then
         stla=12345
         stlo=12345
         evla=12345
         evlo=12345
        endif
        if (itr.gt.1) then
           if (nsamp.gt.nsmp) then
              write(*,*)' WARNING: use only ',nsmp,' samples on'//
     & ' trace ',itr
           endif
           if (abs(dt-dt1).gt.dt1*0.01) then
c            stop ' Traces have different dt ! '
             write(*,*)' ATTENTION: trace ',itr,' has a different dt !'
             write(*,*)' ATTENTION: skipping trace ',itr,' .'
             nskip=nskip+1
             goto 999
           endif
           if (abs(beg1-beg).gt.dt1) then
c            stop ' Traces have different begin time.'
             write(*,*)' ATTENTION: trace ',itr,' has a different beg !'
             write(*,*)' ATTENTION: skipping trace ',itr,' .'
             nskip=nskip+1
             goto 999
           endif
        endif
        call getfhv('depmax',depmax,nerr1)
        if (isnan(depmax)) then
          write(*,*)' ATTENTION: trace ',itr,' with problems (NaNs)'
          write(*,*)' ATTENTION: skipping trace ',itr,' .'
          nskip=nskip+1
          goto 999
        endif

C RM MEAN
c--------
        if (lrm) then
          rmean=0.0
!$OMP PARALLEL default(shared) private (i)
!$OMP DO REDUCTION (+:rmean)
          do i=1,nsamp
            rmean=rmean+sig(i)
          enddo
!$OMP END DO
!$OMP DO
          do i=1,nsamp
            sig(i)=sig(i)-rmean
          enddo
!$OMP END DO
!$OMP END PARALLEL
        endif
c linear t-domain stack
c-----------------------
cc!$OMP PARALLEL DO default(shared)
cc!$OMP& PRIVATE (i)
        do i=1,nsamp
           rsig(i)=sig(i)+rsig(i)
        enddo
cc!$OMP END PARALLEL DO

c get power for FFT:
c--------------------
        if (itr.eq.1) then
          beg1=beg
          dt1=dt
          npow=1
          nsmp=2
 5        if (nsamp.le.nsmp) goto 7
          npow=npow+1
          nsmp=nsmp*2
          goto 5
          stop ' problem with power?'
 7        nsmp2=nsmp/2
c         df=1./(dt*float(nsmp))
        endif
        
c S-transform
c------------
        call st_4ward(sig,nsamp,cycle,dt,ctf,nsmp,npow)

c tf-domain stack ctf2 & PS
c-----------------------
!$OMP PARALLEL DO
!$OMP& PRIVATE (ns,nf,cdum,rdum)
!$OMP& SHARED (ctf2,ctfps,nsmp,nsmp2)
        do ns=1,nsmp
          do nf=1,nsmp2
             cdum=ctf(ns,nf)
             ctf2(ns,nf)=ctf2(ns,nf)+cdum
             rdum=cabs(cdum)
             if (rdum.lt.0.00001) then
               cdum=cmplx(0.,0.)
             else
               cdum=cdum/(cabs(cdum))
             endif
             ctfps(ns,nf)=ctfps(ns,nf)+cdum
          enddo
        enddo
!$OMP END PARALLEL DO
        
 999    continue
        if (itr.le.3) write(*,*)' finished with trace #:',itr
        if (itr.eq.4) write(*,*)'           ...'
        if (mod(itr,20).eq.0) then
           write(*,*)' finished with trace #:',itr
           write(*,*)'           ...'
        endif
c end loop while over all traces!
      enddo
 80   continue
      mtr=itr-1-nskip
      write(*,*)' total # traces       :',mtr

c tf-PWS:
c--------
      rmtr=float(mtr)
!$OMP PARALLEL DO
!$OMP& PRIVATE (ns,nf,cdum,rdum)
!$OMP& SHARED (ctf,ctf2,ctfps,wu,rmtr,nsmp,nsmp2)
      do ns=1,nsmp
        do nf=1,nsmp2
           cdum=ctf2(ns,nf)/rmtr
           rdum=cabs(ctfps(ns,nf))/rmtr
           ctf(ns,nf)=cdum*(rdum**wu)
        enddo
      enddo
!$OMP END PARALLEL DO

c INVERSE TRANSFORM:
c-------------------
      write(*,*)'   ... just before inv transform.'
      if (list) then
       call ist_SML96(ctf,nsamp,nsmp2,npow,dt,sig,stwpi)
      else
       cdum=cmplx(0.0,-twpi/float(nsmp))
       rdum=float(nsmp)*stwpi
!$OMP PARALLEL DO
!$OMP& PRIVATE (ntime,nfre,carg,carray1)
!$OMP& SHARED (rdum,cdum,sig,nsmp,nsmp2,nsamp)
       do ntime=1,nsamp
         carg=cdum*(ntime-1)
         do nfre=1,nsmp2
            carray1(nfre)=ctf(ntime,nfre)*rdum
         enddo
         sig(ntime)=diftc_fast_f(carray1,nsmp,nsmp2,carg)
       enddo
!$OMP END PARALLEL DO
      endif

c Normalize LS:
!$OMP PARALLEL DO
!$OMP& PRIVATE (ntime,nsamp)
!$OMP& SHARED (rsig,rmtr)
      do ntime=1,nsamp
         rsig(ntime)=rsig(ntime)/rmtr 
      enddo
!$OMP END PARALLEL DO

      write(*,*)' ... write stacks: ',tls(1:ntls),' ',tfpws(1:ntfpws)
      if (lkstnm) then
      call wrsac(tls(1:ntls),kstnm,beg,dt,rsig,
     & nsamp,evla,evlo,stla,stlo,luser5,user5,luser6,user6)
      call wrsac(tfpws(1:ntfpws),kstnm,beg,dt,sig,
     & nsamp,evla,evlo,stla,stlo,luser5,user5,luser6,user6)
      else
      call wrsac(tls(1:ntls),'t-lin',beg,dt,rsig,
     & nsamp,evla,evlo,stla,stlo,luser5,user5,luser6,user6)
      call wrsac(tfpws(1:ntfpws),'tf-pws',beg,dt,sig,
     & nsamp,evla,evlo,stla,stlo,luser5,user5,luser6,user6)
      endif

      end
c--------------------------------------------------------------
      SUBROUTINE usage
      write(*,*)'      '
      write(*,*)'USAGE:  tf_pws0MP filename parameter_list '
      write(*,*)'      '
      write(*,*)'filename: One file name (SAC) per line; '
      write(*,*)'          traces must have same b,nsmpl&dt.'
      write(*,*)'      '
      write(*,*)'parameter_list: rm cyc= wu= info kstnm='
      write(*,*)'                 '
      write(*,*)'parameters are optional and can be provided in'
      write(*,*)'arbitrary order without any blank around =.'
      write(*,*)' '
      write(*,*)'rm     : remove mean from input data. '
      write(*,*)'cyc=   : num. of cycles in 2*std (default cyc=2)'
      write(*,*)'wu=    : PWS power (default: wu=2)'
      write(*,*)'kstnm= : optionally specify sac-header (char*8)'
      write(*,*)'user5= : optionally specify sac-header variable user5'
      write(*,*)'user6= : optionally specify sac-header variable user6'
      write(*,*)'tls=   : specify file name for linear stack'
      write(*,*)'         (Default: tl.sac)'
      write(*,*)'tfpws= : specify file name for tf-pws stack'
      write(*,*)'         (Default: tf-pws.sac)'
      write(*,*)'ist    : Use inverse ST (IST) by Stockwell et al.'//
     & ' (1996)'
      write(*,*)'         rather than by Schimmel & Gallart'//
     & ' (2005).'
      write(*,*)'info   : write background and main references'
      write(*,*)'         to screen. Just type: tf_pws0 info'
      write(*,*)
      write(*,*)'EXAMPLE:'
      write(*,*)'Stack all my_data*.sac using default settings:'
      write(*,*)' ls -1 my_data*.sac > list.in'
      write(*,*)' tf_pws0MP list.in '
      write(*,*)
      write(*,*)'DEFAULTS: cyc=2 wu=2 tls=tl.sac tfpws=tf-pws.sac '
      write(*,*)'(Defaults are suited/good for most applications.)'
      write(*,*)' '
      write(*,*)'AUTHOR: Martin Schimmel, (last change 18/04/2017)'
      write(*,*)' '
      write(*,*)'Please, do not hesitate to send bugs, comments'//
     & ', improvements,'
      write(*,*)'nice results, ... to schimmel@ictja.csic.es .'
      write(*,*)' '
      stop
      return
      end
c--------------------------------------------------------------
      SUBROUTINE st_4ward(sig,nsamp,cycle,dt,ctf,nsmp,npow)
c
c INPUT: sig,nsamp,cycle,dt,nsmp,npow
c * sig(nsamp) is real time series with sample interval dt.
c * cycle is the number of periods that equals the 2 std of
c   the Gaussian window.
c * nsmp=2**npow  (==> for FFT)
c * dt is sample interval (time).
c OUTPUT: ctf
c * ctf(nsamp,nsmp2+1) is the complex t-f representation of
c   sig(nsamp).
c   nsamp in ctf is the center sample of 'moving Gauss-window'.
c
c  schimmel@ija.csic.es,  07.05.2004
c  last change 22.08.06
c=================================================================

c     parameter (nsmax=16384,nsctf=16384,nfctf=8193)
      parameter (nsmax=8192,nsctf=8192,nfctf=4097)
c     parameter (nsmax=4096,nfmax=2048)
c nfmax=nsmax/2 + 1, nsctf=2**n, nsctf>nsmax
      real sig(nsmax)
      complex csig(nsctf),ctf(nsctf,nfctf)
      complex cdum(nsctf)
      integer npow,nsmp,nsamp
      real pi,fact,cycle

      pi=4.0*atan(1.0)
      fact=cycle/2.
      nsmp2=nsmp/2

!$OMP PARALLEL default(shared) private (ns)
!$OMP DO
      do ns=1,nsamp
        csig(ns)=cmplx(sig(ns),0.0)
      enddo
!$OMP END DO
!$OMP DO
c to go the safe way:
      do ns=nsamp+1,nsctf
        csig(ns)=cmplx(0.0,0.0)
      enddo
!$OMP END DO
!$OMP END PARALLEL
c
c Perform FFT:
c-------------
      rn=1.
      call clogc(npow,csig,rn,dt)

c=======================
c Start frequency loop:
c=======================
      pp=-2.0*pi*pi*fact*fact

c for n=0 (zero frequency nf=1):
c (zero mean t-series assumed!!!!)
c---------------------------------
!$OMP PARALLEL DO default(shared) private (ns)
      do ns=1,nsamp
        ctf(ns,1)=cmplx(0.0,0.0)
      enddo
!$OMP END PARALLEL DO

c and now remaining positive frequencies:
c----------------------------------------
        rn=-1.
!$OMP PARALLEL DO default(private) 
!$OMP& PRIVATE (nf,nf1,mf,cdum,rpp,rmf,rns,argp,argn,mn,rr,ns)
!$OMP& SHARED (nsmp2,pp,nsmp,ctf,csig,nsamp,rn,dt,npow)
      do nf=2,nsmp2+1
        nf1=nf-1
c zero negative freqs:
        do mf=1,nsmp
          cdum(mf)=cmplx(0.0,0.0)
        enddo
        rpp=pp/float(nf1)/float(nf1)
c       do mf=1,nsmp+1
        do mf=1,nsmp
          rmf=float(mf-1)
          rns=float(nsmp-mf+1)
          argp=rpp*rmf*rmf
          argn=rpp*rns*rns
          if (argp.gt.-25..or.argn.gt.-25.) then
c shift Fourier spectrum csig(mf) by freq nf-1 to csig(mf+nf-1):
            mn=mf+nf1
            if (mn.gt.nsmp) mn=mn-nsmp
c multiply with localizing Gaussian (generalized voice Gaussian):
            rr=exp(argp)+exp(argn)
            cdum(mf)=csig(mn)*rr
          endif
        enddo
         
c perform ifft to obtain temporal spectral localizations:
!$OMP CRITICAL
        call clogc(npow,cdum,rn,dt)
!$OMP END CRITICAL
c store in matrix:
        do ns=1,nsamp
          ctf(ns,nf)=cdum(ns)
        enddo
      enddo
!$OMP END PARALLEL DO

      return
      end
c----------------------------------------------------------------
      SUBROUTINE infooo

      write(*,*)' '
      write(*,*)'tf_pws0.f: Zero-lag PWS in the tf-domain '
      write(*,*)' '
      write(*,*)'This program performs a PWS (phase weighted '//
     & 'stack, Schimmel & '
      write(*,*)'Paulssen, 1997) in the time-frequency domain'//
     & ' employing an'
      write(*,*)'S-Transform (ST). This has been published in'//
     & ' Schimmel & Gallart (2007).'
      write(*,*)'There you find also when and why the ST'//
     & ' is analytic. To go '
      write(*,*)'back into the time domain, we use the inverse'//
     & ' S-transform (IST) from'
      write(*,*)'Schimmel & Gallart (2005). This IST '//
     & 'is also explained '
      write(*,*)'in the appendix of Schimmel & Gallart (2007). '//
     & 'Schimmel et al. (2011)'
      write(*,*)'show how these strategies can be used '//
     & 'with ambient seismic noise.'
      write(*,*)'Informations on how to discretise the S-transform'//
     & ' without side effects'
      write(*,*)'are published in Simon et al. (2007). Further,'//
     & ' Ventosa et al. (2008)'
      write(*,*)'show the relations between the S-transform and'//
     & ' wavelet transform.'
      write(*,*)
      write(*,*)'There exist different ISTs,  e.g., see Ventosa et'//
     & ' al. (2008) and none'
      write(*,*)'of them is perfect. Their limitations are'//
     & ' due to approximations, their'
      write(*,*)'design and discretisation. The discretization is not'//
     & ' trivial and different'
      write(*,*)'artifacts may occur (see Simon et al. 2007, for'//
     & ' details). The IST by Schimmel'
      write(*,*)'and Gallart (2005) has a good time resolution'//
     & ' and was designed for'
      write(*,*)'signal detection and signal enhancement (e.g.,'//
     & ' in polarization and phase'
      write(*,*)'coherence filters). At the lowermost frequency '//
     & 'border frequency sampling '
      write(*,*)'might not be good enough for certain applications'//
     & ' and a logarithmic sampling'
      write(*,*)'of the frequency would have been better (Ventosa et'//
     & ' al. (2008). E.g., the'
      write(*,*)'Gaussian window in the time domain does not imply'//
     & ' a Gaussian in the frequency'
      write(*,*)'domain (at the sparsely sampled low-frequency border'//
     & '). The constant'
      write(*,*)'sampling might be too croase for an artifact-free'//
     & ' waveform reconstruction'
      write(*,*)'at lowest frequency.'
      write(*,*)' '
      write(*,*)'   Schimmel M., and H. Paulssen, Noise reduction '//
     & 'and'
      write(*,*)'     detection of weak, coherent signals through '
      write(*,*)'     phase weighted stacks, GJI, 130, 497-505, 1997.'
      write(*,*)'   Schimmel M., and J. Gallart, The inverse S '//
     & 'Transform in'
      write(*,*)'     filters with time-frequency localization , '//
     & 'IEEE'
      write(*,*)'     Transactions on Signal Processing, 53 (11), '//
     & '4417-4422,'
      write(*,*)'     doi:10.1109/TSP.2005.857065, 2005.'
      write(*,*)'   Schimmel M., and J. Gallart, Frequency-dependent '//
     & 'phase'
      write(*,*)'     coherence for noise suppression in seismic '//
     & 'array data,'
      write(*,*)'     J. Geophys. Res., 112, B04303, '//
     & 'doi:10.1029/2006JB004680, 2007.'
      write(*,*)'   Schimmel, M., Stutzmann, E., Gallart, J., Using '//
     & 'instantaneous '
      write(*,*)'     phase coherence for signal extraction from '//
     & 'ambient noise'
      write(*,*)'     data at a local to a global scale, Geophys. J. '//
     & 'Int., 184, '
      write(*,*)'     494-506, doi:10.1111/j.1365-246X.2010.04861.x, '//
     & '2011.'
      write(*,*)'   Simon et al., The S-transform and its inverses:'//
     & ' side effects '
      write(*,*)'     of discretising and filtering, IEEE '//
     & 'Transactions on Signal'
      write(*,*)'     Processing, 55, 4928-4937, '//
     & 'doi:10.1109/TSP.2007.897893, 2007.'
      write(*,*)'   Ventosa et al., S-transform from a wavelets point'//
     & ' of view,'
      write(*,*)'     IEEE Transactions on Signal Processing, 56, '//
     & '2771-2780,'
      write(*,*)'     doi:10.1109/TSP.2008.917029, 2008.'
      write(*,*)
      write(*,*)' Author: Martin Schimmel (schimmel@ija.csic.es)'
      write(*,*)

      return
      end
c--------------------------------------------------------------------
      SUBROUTINE wrsac(name,kst,b,dt,y,nsmpl,
     & evla,evlo,stla,stlo,luser5,user5,luser6,user6) 

      real y(1),b,dt,gcarc,evdp,baz,user5,user6
      integer nsmpl
      character*(*) kst,name
      logical luser5,luser6

 
      call newhdr
      call setnhv('npts',nsmpl,nerr)
      call setfhv('delta',dt,nerr)
      call setkhv('kstnm',kst,nerr)
      if (luser5) call setfhv('user5',user5,nerr)
      if (luser6) call setfhv('user6',user6,nerr)
      call setfhv('evla',evla,nerr)
      call setfhv('evlo',evlo,nerr)
      call setfhv('stla',stla,nerr)
      call setfhv('stlo',stlo,nerr)
 
c     e=b+(nsmpl-1)*dt
c     o=0.
 
      call setfhv('b',b,nerr)
c     call setihv('iztype','io',nerr)
c
      call wsac0(name,dum,y,nerr)
      if(nerr.ne.0) stop 'Error in writing output file'
 
      return
      end
c--------------------------------------------------------------
      SUBROUTINE CLOGC(n,x,sigm,dt)
 
c--- performs fft on signals with length 2**n and sampling interval
c--- of dt seconds (if in the time domain; notice that dt*df=1/2**n).
c--- the signal is stored in x. it may be complex.
c--- the spectrum is returned in x. it is almost always complex.
c--- a time-to-frequency transform is done with sign=+1. (conform
c--- the convention adopted in aki and richards - the alternative
c--- convention may be obtained by taking complex conjugates after
c--- the call to clogc).
c--- the normalization factor 1./twopi occurs in the frequency-to
c--- time transform (again aki&richards).
c--- normalization is such that physical dimensions are respected.
c--- thus, if the time signal is dimensioned in meters, the
c--- resulting spectral density in x is in meters/hz. for example,
c--- if the time signal is the unit sinc function of width dt, centered
c--- at t=0, the spectral density is dt for all values of the frequency.c
c--- array locations: if x contains the spectrum, it has the spectrum
c--- for positive frequencies in the first 2**n/2+1 elements, such that
c--- x(1) is at 0 hz, x(2) at df hertz, and x(2**n/2+1) at the nyquist,
c--- where df=1./(2**n*dt) and the nyquist is 1./(2*dt) hz.
c--- the second half of x contains the spectrum for negative frequencies
c--- such that x(2**n) is at -df, x(2**n-1) at -2*df hz etcetera.
c--- if x contains the time signal, x(1) is at time 0, x(2)
c--- at time dt etc.
c
      dimension x(1),m(25)
      complex x,wk,hold,q
      if(sigm.ge.0.) then
        sign=1.
      else
        sign=-1.
      endif
      lx=2**n
      do 1 i=1,n
    1 m(i)=2**(n-i)
      do 4 l=1,n
      nblock=2**(l-1)
      lblock=lx/nblock
      lbhalf=lblock/2
      k=0
      do 4 iblock=1,nblock
      fk=k
      flx=lx
      v=sign*6.283185308*fk/flx
      wk=cmplx(cos(v),sin(v))
      istart=lblock*(iblock-1)
      do 2 i=1,lbhalf
      j=istart+i
      jh=j+lbhalf
      q=x(jh)*wk
      x(jh)=x(j)-q
      x(j)=x(j)+q
    2 continue
      do 3 i=2,n
      ii=i
      if(k.lt.m(i)) go to 4
    3 k=k-m(i)
    4 k=k+m(ii)
      k=0
      do 7 j=1,lx
      if(k.lt.j) go to 5
      hold=x(j)
      x(j)=x(k+1)
      x(k+1)=hold
    5 do 6 i=1,n
      ii=i
      if(k.lt.m(i)) go to 7
    6 k=k-m(i)
    7 k=k+m(ii)
      if(sign.gt.0.) go to 9
      flx=flx*dt
      do 8 i=1,lx
    8 x(i)=x(i)/flx
      return
    9 do 10 i=1,lx
   10 x(i)=x(i)*dt
      return
      end 
c------------------------------------------------------------------
      SUBROUTINE ist_SML96(ctf,nsamp,nsmp2,npow,dt,sig,r)
c
c Compute inverse S-transform following 
c Stockwell, Mansinha & Lowe, 1996, IEEE Trans. Signal Proc., 44,
c
c INPUT:
c * ctf(nsmax,nfmax): complex tf-matrix (S-spectrum).
c * nsamp: number of samples t-series.
c * npow: 2**npow >= nsamp 
c * nsmp2: nsmp2=2**(npow-1)
c * dt: sample intervall.
c
c OUTPUT:
c * sig(nsmax): real time series, nsamp samples.
c
c NOTES:
c Please, send bugs, comments, improvements, other routines
c to schimmel@ija.csic.es . 
c_____________________________________________________________________
c
      parameter (nsmax=8192,nfmax=4096)
      complex csig(nsmax),ctf(nsmax,nfmax)
      complex cdum
      real sig(nsmax),dt,r
      integer nsmpl2,npow,nsamp

c Average the voices of S(n/NT,jT):
c----------------------------------
      do nf=1,nsmp2
        cdum=cmplx(0.0,0.0)
        do ns=1,nsamp
          cdum=cdum+ctf(ns,nf)
        enddo
        csig(nf)=cdum
      enddo
      do nf=2,nsmp2
        ii=nsmp-nf+2
        cdum=conjg(csig(nf))
        csig(ii)=cdum
      enddo
      csig(1)=cmplx(0.,0.)
      csig(nsmp2+1)=cmplx(0.,0.)
c
c Perform IFFT:
c-------------
      rn=-1.0
      call clogc(npow,csig,rn,dt)

c Get real time series:
c----------------------
      do ns=1,nsamp
         sig(ns)=real(csig(ns))*dt*r
      enddo

      return
      end
c-------------------------------------------------------------
      integer FUNCTION leng(char)
c     leng is the position of the last non blank character
c     in the character variable char
      character char*(*)
      integer*4 l
      l=len(char)
      do 1 leng=l,1,-1
    1 if (char(leng:leng) .ne. ' ') return
      leng=0
      return
      end   
c-------------------------------------------------------------
        REAL FUNCTION diftc_fast_f(y,nsmp,nsmp2,carg)
c
c Discrete Inverse Fourier Transformation / |f|:
c (Note: constant factors are not included.)
c
c  y        is array with complex spectrum (positive freqs).
c  nsmp     is # of samples in time domain.
c  nsmp2    is the length of data vector y (pos. freqs).
c  ntime    is desired time index.
c  carg     is cmplx(0.0,twpi*ntime/nsmp)
c===============================================================
        complex y(1),carg,yy,cdum,cphs1,cphs2
        integer nsmp2,nsmp

        cdum=cmplx(0.D0,0.D0)
c       do nm=1,nsmp2
c          yy=y(nm)/float(nm)

        do nm=2,nsmp2
           yy=y(nm)/float(nm-1)
           cphs1=yy*cexp(carg*(nm-1))
           cphs2=conjg(yy)*cexp(carg*(nsmp-nm+1))
           cdum=cdum+cphs1+cphs2
        enddo
        diftc_fast_f=real(cdum)/float(nsmp)

       end
c---------------------------------------------------------------
